from __future__ import annotations

import os.path
from contextlib import closing
from typing import Container

from moviepy.video.io.VideoFileClip import VideoFileClip
from tqdm import tqdm

from FastDub import Subtitles, Voicer, Audio
from FastDub.FFMpeg import FFMpegWrapper

__all__ = ('Dubber',)


class Dubber:
    __slots__ = (
        'fit_align', 'language',
        'ducking', 'min_silence_len', 'silence_thresh',
        'gain_during_overlay'
    )

    VOISER = Voicer.Voicer()

    def __init__(self, voice: str, language: str, ducking: bool,
                 min_silence_len: int, silence_thresh: float, gain_during_overlay: int,
                 fit_align: float = 2.):

        self.language = language
        self.fit_align = fit_align
        if voice:
            self.VOISER.set_voice(voice)

        self.ducking = ducking
        self.min_silence_len = min_silence_len
        self.silence_thresh = silence_thresh
        self.gain_during_overlay = gain_during_overlay

    @staticmethod
    def collect_videos(path_to_files: str, skip_starts_underscore: bool = True,
                       exclude_files: Container[str] = None):
        path_to_files = os.path.abspath(path_to_files)
        videos = {}
        for file in os.listdir(path_to_files):
            if file in exclude_files:
                continue
            absfile = os.path.join(path_to_files, file)
            if not os.path.isfile(absfile) or skip_starts_underscore and file.startswith('_'):
                continue
            filename, ext = os.path.splitext(file)
            if videos.get(filename):
                videos[filename][ext] = absfile
                continue
            videos[filename] = {ext: os.path.join(path_to_files, file)}
        return videos

    def dub_dir(self, videos: dict[str, dict[str, str]], video_format: str, subtitles_format: str):
        for fn, exts in videos.items():
            self.dub_one(fn, exts.get(video_format), exts.get(subtitles_format))

    def dub_one(self, fn: str, target_mp4: str, target_srt: str, clean_up: int = 1):
        if target_mp4 is target_srt is None:
            return
        subtitles = Subtitles.parse(target_srt)

        default_right_border = 0
        if target_mp4:
            video_clip: VideoFileClip
            with closing(VideoFileClip(target_mp4, audio=False)) as video_clip:
                default_right_border = video_clip.end * 1000. - subtitles[-1].ms.end
        default_right_border = max((default_right_border, 0))

        tqdm_total = len(subtitles)

        tts_list = [self.VOISER.voice(line.text) for line in
                    tqdm(subtitles, desc='TTSing', total=tqdm_total,
                         unit='line', dynamic_ncols=True, colour='white')]

        end_element_pos = tqdm_total - 1
        audio = Audio.AudioSegment.empty()

        for pos, tts, line in tqdm(
                ((i, tts, subtitles[i]) for i, tts in enumerate(tts_list)),
                desc='Fitting Audio',
                total=tqdm_total, unit='line',
                dynamic_ncols=True, colour='white'):
            audio += Audio.fit(
                tts,
                line.ms.start - audio.duration_ms,
                line.ms.duration,
                (default_right_border
                 if pos == end_element_pos else
                 (subtitles[
                      pos + 1].ms.start - line.ms.end)
                 ), self.fit_align
            )

        max_duration = subtitles[-1].ms.end
        audio_duration = audio.duration_ms

        if audio_duration > max_duration:
            change_speed = audio_duration / max_duration
            print(f'Changing audio speed to {round(change_speed, 3)}')
            audio = Audio.speed_change(audio, change_speed)
        elif audio_duration != max_duration:
            audio = Audio.AudioSegment.silent(
                min((max_duration - audio_duration, subtitles[0].ms.start))
            ) + audio

        result_dir = os.path.join(os.path.dirname(target_srt), f'_{fn}_result')
        if not os.path.isdir(result_dir):
            os.mkdir(result_dir)

        target_out_audio = os.path.join(result_dir, f'_{self.language}.mp3')
        result_out_audio = os.path.join(result_dir, f'{self.language}.mp3')

        if target_mp4:
            with closing(VideoFileClip(target_mp4)) as video_clip:
                video_clip.audio.write_audiofile(target_out_audio)
                if self.ducking:
                    Audio.side_chain(Audio.AudioSegment.from_file(target_out_audio), audio, self.min_silence_len,
                                     self.silence_thresh, self.gain_during_overlay
                                     ).export(result_out_audio)
                else:
                    Audio.AudioSegment.from_file(target_out_audio).overlay(audio, self.gain_during_overlay
                                                                           ).export(result_out_audio)

            FFMpegWrapper.replace_audio_in_video(target_mp4, result_out_audio,
                                                 os.path.join(result_dir, f'{self.language}.mp4'))
            if clean_up > 0:
                os.remove(target_out_audio)
                if clean_up > 1:
                    os.remove(result_out_audio)
                    if clean_up > 2:
                        self.VOISER.cleanup()
        else:
            audio.export(result_out_audio)



try:
    import download_youtube_subtitle.main
    from FastDub.YT import Subtitles
    from FastDub.YT import pafy
    from FastDub.YT.pafy import g
    from FastDub.YT.pafy.backend_youtube_dl import YtdlPafy
except ImportError as e:
    from typing import Any as YtdlPafy
    SUPPORTED = False
else:
    SUPPORTED = True



import multiprocessing.pool
from pathlib import Path
from shutil import get_terminal_size
from typing import Callable, TypeVar, Optional, Iterable
from urllib.parse import urlparse

from FastDub.YT import *

__all__ = 'DownloadYTVideo',
_API_RET_TYPE = TypeVar('_API_RET_TYPE')


class DownloadYTVideo:
    __slots__ = ('save_dir', 'language', 'playlist', 'API_KEYS')

    def __init__(self, url: str, language: str, api_keys: Iterable[str] = ()):
        self.API_KEYS = {g.api_key, 'AIzaSyCHxJ84-ryessLJfWZVWldiuVCnxtf0Nm4', *api_keys}

        if urlparse(url).path == '/playlist':
            playlist = self.with_api_key(lambda: pafy.get_playlist2(url))
            save_dir = playlist.plid
        else:
            playlist = self.with_api_key(lambda: (pafy.new(url),))
            save_dir = playlist[0].videoid
        save_dir = Path(save_dir)
        if not save_dir.is_dir():
            save_dir.mkdir()

        self.save_dir = save_dir
        self.language = language
        self.playlist = playlist

    def multiprocessing_download(self, pc: int = None):
        if pc is None:
            pc = multiprocessing.cpu_count()
        if pc < 2:
            for yt_dl in self.playlist:
                self.download(yt_dl)
            return
        with multiprocessing.pool.ThreadPool(pc) as pool:
            pool.map(self.download, self.playlist)

    def download(self, yt_dl: YtdlPafy):
        save_to = self.save_dir / yt_dl.videoid
        srt_file = Path(f'{save_to}.srt')
        if not srt_file.is_file():
            Subtitles.download_srt(yt_dl.videoid, self.language, srt_file)
        mp4_file = Path(f'{save_to}.mp4')
        if not mp4_file.is_file():
            try:
                self.with_api_key(self.mp4_downloader(yt_dl, mp4_file))
                print(f'\r{mp4_file} downloaded.'.ljust(get_terminal_size().columns))
            except OSError:
                pass

    def with_api_key(self, func: Callable[[], _API_RET_TYPE]) -> Optional[_API_RET_TYPE]:
        for key in self.API_KEYS:
            pafy.set_api_key(key)
            try:
                return func()
            except pafy.util.GdataError:
                continue

    def mp4_downloader(self, yt_dl: YtdlPafy, mp4_file: str | Path):
        return lambda: yt_dl.getbest('mp4').download(f'{mp4_file}',
                                                     progress='MB',
                                                     callback=self.progress_callback)

    @staticmethod
    def progress_callback(total: int, downloaded: float, ratio: float, rate: float, eta: float):
        print(
            end=f'\r[{ratio:.2%}] {downloaded:,.2f}/{total/1048576:,.2f}MB. {rate:,.2f} kb/s: '
                f'ETA {eta} sec.'
                ''.ljust(get_terminal_size().columns),
            flush=True)



from __future__ import annotations

import argparse
import multiprocessing
import os
from math import inf
from time import perf_counter

from FastDub import Dubber, Voicer
from FastDub import Translator
from FastDub import YT
from FastDub.FFMpeg import FFMpegWrapper
from FastDub.Translator.SubtitlesTranslate import SrtTranslate
from FastDub.YT.Downloader import DownloadYTVideo


def parse_args() -> argparse.Namespace:
    arg_parser = argparse.ArgumentParser('FastDub',
                                         description='FastDub is a tool for dubbing videos by subtitle files.',
                                         formatter_class=argparse.RawTextHelpFormatter)

    arg_parser.add_argument('-rc', '--remove-cache', default=0, type=int, choices=(0, 1, 2),
                            help='Remove all cache (_cached_texts directory) files\n'
                                 '\t0 - No remove cache (default)\n'
                                 '\t1 - Delete cache before voice acting\n'
                                 '\t2 - Delete cache after voice acting')
    arg_parser.add_argument('-rf', '--cleanup-level', default=1,
                            help='Cleanup level'
                                 '\t0 = No removing any files\n'
                                 '\t> 0 remove audio from video (default)\n'
                                 '\t> 1 = remove dubbed audio if video exists\n'
                                 '\t> 2 = reomve dubbed cache files')

    arg_parser.add_argument('-l', '--language', default='ru',
                            help='Subtitles language (ru)')

    if YT.SUPPORTED or Translator.SUPPORTED:
        arg_parser.add_argument('-tc', '--threads-count', default=multiprocessing.cpu_count(),
                                type=lambda i: int(multiprocessing.cpu_count() * float(i.lstrip('*')))
                                if i.startswith('*') else int(i),
                                help='Process count to download (pass to cpu count, < 2 to disable)\n'
                                     '\t*N = N * cpu count')

    input_group = arg_parser.add_argument_group('Input')
    input_group.add_argument('-i', '--input', type=str, default=os.getcwd(),
                             help=f"Input directory{('/YouTube Playlist/Video URL' if YT.SUPPORTED else '')}.")
    input_group.add_argument('-vf', '--video-format', default='.mp4',
                             help='Video format (default .mp4).')
    input_group.add_argument('-sf', '--subtitles-format', default='.srt',
                             help='Subtitles format (default .srt).')

    exclude_input_group = input_group.add_argument_group('Exclude Input')
    exclude_input_group.add_argument('-En', '--exclude', nargs='+', default=(),
                                     help='Exclude <name>')
    exclude_input_group.add_argument('-Eu', '--exclude-underscore', default=True,
                                     help='Exclude files starts with underscore')

    ducking_group = arg_parser.add_argument_group('Audio Ducking')
    ducking_group.add_argument('-sc', '--sidechain', action=argparse.BooleanOptionalAction, default=True,
                               help='Enable audio side chain compress (ducking)')
    ducking_group.add_argument('-sc-msl', '--min-silence-len', '--attack', default=100, type=int,
                               help='Minimum silence length in ms (default 100)')
    ducking_group.add_argument('-sc-st', '--silence-thresh', default=-inf, type=float,
                               help='Silence threshold in dB')
    ducking_group.add_argument('-sc-gdo', '--gain-during-overlay', default=-11, type=int,
                               help='Gain during overlay in dB (-11)')

    voicer_group = arg_parser.add_argument_group('Voicer')
    voicer_group.add_argument('-v', '--voice', type=str.lower, choices=Voicer.VOICES_NAMES.keys(),
                              help='SAPI voice for voice acting.')
    voicer_group.add_argument('-a', '--align', default=2., type=float,
                              help='Audio fit align\n'
                                   '\t1 = right\n'
                                   '\t2 = center (default)')

    ffmpeg_group = arg_parser.add_argument_group('FFMpeg Output')
    ffmpeg_group.add_argument('-ll', '--loglevel', default='panic',
                              help='FFMpegWrapper loglevel')
    ffmpeg_group.add_argument('-y', '--confirm', action=argparse.BooleanOptionalAction, default=True,
                              help="Don't ask for confirmation")

    if YT.SUPPORTED:
        yt_group = arg_parser.add_argument_group('Youtube')
        yt_group.add_argument('-yt', '--youtube', action='store_true', default=False)

        yt_group.add_argument('-ak', '--api-keys', nargs='+', default=(), help='Youtube API key/s')

    if Translator.SUPPORTED:
        translate_group = arg_parser.add_argument_group('Translate subtitles')
        translate_group.add_argument('-tr', '--translate', action='store_true', default=False,
                                     help='Translate input subtitles files')
        translate_group.add_argument('--rewrite-srt', action=argparse.BooleanOptionalAction, default=False,
                                     help='Rewrite input subtitles files.\n'
                                          'If not, add "_" to the beginning of the original subtitle file.')
        translate_group.add_argument('-ts', '--translate-service', type=str, default='google',
                                     choices=[i[1:] for i in dir(Translator.translators.apis) if
                                              i.startswith('_') and not i.startswith('__')],
                                     help='Subtitle translation service. (default google)')

    return arg_parser.parse_args()


def main():
    args = parse_args()
    remove_cache = args.remove_cache
    if remove_cache == 1:
        Voicer.Voicer().cleanup()

    FFMpegWrapper.DEFAULT_FFMPEG_LOG_LEVEL = args.loglevel

    total_time = 0.
    if args.confirm:
        FFMpegWrapper.DEFAULT_ARGS += '-y',
        total_time = perf_counter()

    if YT.SUPPORTED and args.youtube:
        downloader = DownloadYTVideo(args.input, args.language, args.api_keys)
        downloader.multiprocessing_download(args.threads_count)
        args.input = downloader.save_dir

    video_format = video_format if (video_format := args.video_format).startswith('.') else f'.{video_format}'
    subtitles_format = (
        subtitles_format if (subtitles_format := args.subtitles_format).startswith('.') else f'.{subtitles_format}')

    videos = Dubber.Dubber.collect_videos(args.input, args.exclude_underscore, args.exclude)

    if Translator.SUPPORTED and args.translate:
        SrtTranslate(args.language, args.translate_service, args.rewrite_srt, args.threads_count
                     ).translate_dir(videos, subtitles_format)

    dubber = Dubber.Dubber(args.voice, args.language, args.sidechain, args.min_silence_len, args.silence_thresh,
                           args.gain_during_overlay, args.align)

    dubber.dub_dir(videos, video_format, subtitles_format)

    if remove_cache == 2:
        Voicer.Voicer().cleanup()

    if total_time:
        print(f'Total time: {perf_counter() - total_time:,.3f} s.')


if __name__ == '__main__':
    main()



from __future__ import annotations

import os.path
from array import array
from copy import copy
from math import inf
from tempfile import TemporaryDirectory

import pydub
import pydub.silence
from tqdm import tqdm

from FastDub.FFMpeg import FFMpegWrapper

__all__ = ('AudioSegment', 'speed_change', 'fit', 'side_chain')


class AudioSegment(pydub.AudioSegment):
    __slots__ = ('duration_ms',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.duration_ms = self.duration_seconds * 1000.

    def append(self, seg, _=None):
        """pydub.AudioSegment. Without crossfade."""
        seg1, seg2 = AudioSegment._sync(self, seg)
        return seg1._spawn(seg1._data + seg2._data)

    def __add__(self, other):
        """Add, only for AudioSegments"""
        return self.append(other)


def speed_change(audio: AudioSegment, speed_changes: float, allow_copy: bool = True, log_level: str = 'trace'
                 ) -> AudioSegment:
    if speed_changes == 1.:
        return audio if allow_copy else copy(audio)
    if speed_changes <= 0.:
        raise ValueError(f"Speed cannot be negative ({speed_changes}).")

    atempo = array('d')
    if speed_changes < .5:
        _n = speed_changes
        while _n < .5:
            atempo.append(.5)
            _n *= 2.
        atempo.append(_n)
    elif speed_changes > 100.:
        _n = speed_changes
        while _n > 100.:
            atempo.append(100.)
            _n /= 2.
        atempo.append(_n)
    else:
        atempo.append(speed_changes)

    with TemporaryDirectory() as tmp:
        inp = os.path.join(tmp, 'inp.mp3')
        out = os.path.join(tmp, 'out.mp3')
        audio.export(inp)
        FFMpegWrapper.convert('-i', inp,
                              '-af', ','.join([f'atempo={i}' for i in atempo]),
                              out, loglevel=log_level)
        return AudioSegment.from_file(out)


def fit(audio: AudioSegment, left_border: float, need_duration: float, right_border: float,
        align: float) -> AudioSegment:
    """Fits audio to the borders of the subtitles."""
    if (audio_duration := audio.duration_ms) > (free := left_border + need_duration + right_border):
        audio = speed_change(audio, audio_duration / free)
        audio_duration = audio.duration_ms
    return AudioSegment.silent(((free - audio_duration) / align)
                               if (audio_duration > (need_duration + right_border))
                               else left_border) + audio


def side_chain(sound1: AudioSegment, sound2: AudioSegment,
               min_silence_len: int = 100, silence_thresh: float = -inf,
               gain_during_overlay: int = -10
               ) -> AudioSegment:
    for start, end in tqdm(
            pydub.silence.detect_nonsilent(sound2, min_silence_len=min_silence_len, silence_thresh=silence_thresh),
            desc="Ducking", colour='white'
    ):
        sound1 = sound1.overlay(sound2[start:end], position=start, gain_during_overlay=gain_during_overlay)
    return sound1



from __future__ import annotations

import argparse

if not hasattr(argparse, 'BooleanOptionalAction'):

    class BooleanOptionalAction(argparse.Action):
        def __init__(self, option_strings, dest, default=None, type=None,
                     choices=None, required=False, help=None, metavar=None):
            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)
                if option_string.startswith('--'):
                    option_string = '--no-' + option_string[2:]
                    _option_strings.append(option_string)
            if help is not None and default is not None and default is not argparse.SUPPRESS:
                help += " (default: %(default)s)"
            super().__init__(option_strings=_option_strings, dest=dest, nargs=0, default=default, type=type,
                             choices=choices, required=required, help=help, metavar=metavar)

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith('--no-'))

        def format_usage(self):
            return ' | '.join(self.option_strings)


    argparse.BooleanOptionalAction = BooleanOptionalAction



import json
import os.path
import time
from math import modf
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable

from download_youtube_subtitle.main import main

__all__ = ('download_srt',)


def _float_to_srt_time_format(d: float) -> str:
    fraction, whole = modf(d)
    return f"{time.strftime('%H:%M:%S,', time.gmtime(whole))}{f'{fraction:.3f}'.replace('0.', '')}"


def _from_json(translation: Iterable[dict[str, str]]) -> str:
    srt_s = ''
    for i, el in enumerate(translation, 1):
        srt_s += (f"\n\n{i}\n{_float_to_srt_time_format((start := int(el['start'])) / 1000.)}"
                  " --> "
                  f"{_float_to_srt_time_format((start + int(el['dur'])) / 1000.)}\n{el['text']}")
    return srt_s.lstrip('\n')


def download_srt(video_id: str, lang: str, fp: str | Path):
    with TemporaryDirectory() as tmp:
        tmp_name = os.path.join(tmp, 'target.json')
        main(video_id, lang, output_file=tmp_name, to_json=True)
        with open(tmp_name, encoding='UTF-8') as f:
            _json: dict[str, list[dict[str, str]]] = json.load(f)
    with open(fp, 'w', encoding='UTF-8') as srt_f:
        srt_f.write(_from_json(_json.get('translation', ())))



import setuptools

with open('requires.txt', encoding='UTF-8') as f:
    requires = f.read().strip().splitlines()
with open('yt_requires.txt', encoding='UTF-8') as f:
    yt_requires = f.read().strip().splitlines()
with open('tr_requires.txt', encoding='UTF-8') as f:
    tr_requires = f.read().strip().splitlines()

with open('README.md', encoding='UTF-8') as f:
    readme = f.read()

setuptools.setup(
    name="PyFastDub",
    description="A Python package (+CLI) for voice over subtitles, with the ability to embed in video, audio ducking, "
                "and dynamic voice changer for a single track.",
    long_description=readme,
    long_description_content_type="text/markdown",
    version="2.4.7",
    author="Nikita (NIKDISSV)",
    install_requires=requires,
    extras_require={'YT': yt_requires, 'TR': tr_requires},
    author_email='nikdissv@proton.me',
    project_urls={
        'Download Voices': 'https://rhvoice.su/voices/',
        'GitHub': 'https://github.com/NIKDISSV-Forever/FastDub'
    },
    packages=setuptools.find_packages(),
    license='MIT',
    python_requires='>=3.8<3.11',
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'License :: OSI Approved',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',

        'Topic :: Multimedia',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Multimedia :: Sound/Audio :: Conversion',
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Utilities',

        'Typing :: Typed',

    ]
)



from __future__ import annotations

from subprocess import check_call

from imageio_ffmpeg import get_ffmpeg_exe

__all__ = ('FFMpegWrapper',)


class FFMpegWrapper:
    __slots__ = ()
    DEFAULT_FFMPEG_LOG_LEVEL = 'panic'
    DEFAULT_ARGS = ()
    FFMPEG_EXE = get_ffmpeg_exe()

    @classmethod
    def convert(cls, *args, loglevel=None):
        if not loglevel:
            loglevel = cls.DEFAULT_FFMPEG_LOG_LEVEL
        check_call([cls.FFMPEG_EXE, '-v', loglevel, *cls.DEFAULT_ARGS, *args])

    @staticmethod
    def replace_audio_in_video(video_path: str, audio_path: str, output_path: str):
        FFMpegWrapper.convert('-i', video_path, '-i', audio_path,
                              '-map', '0:0', '-map', '1:0', '-c', 'copy',
                              output_path)
